from collections.abc import Iterator
from contextlib import contextmanager
from enum import Enum
from pathlib import Path

from playwright.sync_api import BrowserContext
from playwright.sync_api import sync_playwright

from ssb_pubmd.request_handler import Response


class CreateContextMethod(Enum):
    """The method used to create the browser context.

    Can be either from a file containing the context data,
        or from a login popup window.
    """

    FROM_FILE = "from_file"
    FROM_LOGIN = "from_login"


class BrowserRequestHandler:
    """This class is used to create a logged in browser context from which to send requests."""

    def __init__(self, context_file_path: Path, login_url: str) -> None:
        """Initializes an empty browser context object."""
        self._context_file_path: Path = context_file_path
        self._login_url: str = login_url
        self._context: BrowserContext | None = None

    @contextmanager
    def new_context(
        self, method: CreateContextMethod = CreateContextMethod.FROM_FILE
    ) -> Iterator[BrowserContext]:
        """Wrapper around playwright's context manager.

        The default is to create a new context from a file.
        If `from_file` is  set False, a new context is created through a browser popup with user login,
            and the context is saved to a file.
        """
        with sync_playwright() as playwright:
            browser = playwright.chromium.launch(headless=False)
            match method:
                case CreateContextMethod.FROM_FILE:
                    self._context = browser.new_context(
                        storage_state=self._context_file_path
                    )
                case CreateContextMethod.FROM_LOGIN:
                    self._context = browser.new_context()
                    login_page = self._context.new_page()
                    login_page.goto(self._login_url)
                    login_page.wait_for_event("close", timeout=0)
                    self._context.storage_state(path=self._context_file_path)
            yield self._context
            self._context.close()
            browser.close()

    def send_request(
        self,
        url: str,
        headers: dict[str, str] | None = None,
        data: dict[str, str] | None = None,
    ) -> Response:
        """Sends a request to the specified url, optionally with headers and data, within the browser context."""
        if self._context is None:
            raise ValueError("Browser context has not been created.")

        api_response = self._context.request.post(
            url,
            data=data,
        )

        try:
            body = api_response.json()
            body = dict(body)
        except Exception:
            body = {}

        response = Response(
            status_code=api_response.status,
            body=body,
        )

        return response



import json
from pathlib import Path

import nbformat
from nbformat import NotebookNode

from ssb_pubmd.constants import METADATA_FILE
from ssb_pubmd.constants import ContentType
from ssb_pubmd.request_handler import RequestHandler
from ssb_pubmd.request_handler import Response


class MarkdownSyncer:
    """This class syncs a content file to a CMS (Content Management System).

    The CMS must have an endpoint that satisfies the following constraints:

    -   It must accept a post request with fields *_id*, *displayName* and *markdown*.
    -   The response body must have a key *_id* whose value should be
        a unique string identifier of the content.

    Creating and updating content is handled in the following way:

    -   On the first request, an empty string is sent as *_id*.
    -   If the request succeeds, the value of *_id* (in the response) is stored in a JSON file
        (created in the same directory as the markdown/notebook file).
    -   On subsequent requests, the stored value is sent as *_id*.
    """

    ID_KEY = "_id"

    def __init__(
        self,
        post_url: str,
        request_handler: RequestHandler,
        metadata_file: Path = METADATA_FILE,
    ) -> None:
        """Creates a markdown syncer instance that connects to the CMS through the post url."""
        self._post_url: str = post_url
        self._request_handler: RequestHandler = request_handler
        self._content_file_path: Path = Path()
        self._content_file_type: ContentType = ContentType.MARKDOWN
        self._metadata_file_path: Path = metadata_file

    @property
    def content_file_path(self) -> Path:
        """Returns the path of the content file."""
        return self._content_file_path

    @content_file_path.setter
    def content_file_path(self, path: Path) -> None:
        """Sets the path of the content file."""
        if not path.is_file():
            raise FileNotFoundError(f"The file {path} does not exist.")

        ext = path.suffix.lower()
        for t in ContentType:
            if ext == t.value:
                self._content_file_type = t
                break
        else:
            allowed_extensions = [t.value for t in ContentType]
            sep = ", "
            raise ValueError(
                f"The file {path} has extension {ext}, but should be one of: {sep.join(allowed_extensions)}."
            )

        self._content_file_path = path

    @property
    def basename(self) -> str:
        """The name of the content file without extension."""
        return self._content_file_path.stem

    @property
    def display_name(self) -> str:
        """Generate a display name for the content."""
        return self.basename.replace("_", " ").title()

    @property
    def metadata_file_path(self) -> Path:
        """The path of the metadata file."""
        return self._metadata_file_path

    @property
    def metadata_key(self) -> str:
        """The key that the content metadata will be stored under in the metadata file."""
        return str(self._content_file_path.absolute())

    def _save_content_id(self, content_id: str) -> None:
        """Saves the content id to the metadata file."""
        with open(self._metadata_file_path) as f:
            try:
                data = json.load(f)
            except json.JSONDecodeError:
                data = {}

        data[self.metadata_key] = {
            self.ID_KEY: content_id,
        }

        with open(self._metadata_file_path, "w") as f:
            json.dump(data, f, indent=4)

    def _get_content_id(self) -> str:
        """Fetches the content id from the metadata file if it exists, otherwise an empty string."""
        with open(self._metadata_file_path) as f:
            try:
                data = json.load(f)
            except json.JSONDecodeError:
                data = {}

        metadata: dict[str, str] = data.get(self.metadata_key, {})

        content_id = metadata.get(self.ID_KEY, "")

        return content_id

    def _read_notebook(self) -> NotebookNode:
        """Reads the notebook file and returns its content."""
        return nbformat.read(self._content_file_path, as_version=nbformat.NO_CONVERT)  # type: ignore

    def _get_content_from_notebook_file(self) -> str:
        """Extracts all markdown cells from the notebook and returns it as a merged string."""
        notebook = self._read_notebook()

        markdown_cells = []
        for cell in notebook.cells:
            if cell.cell_type == "markdown":
                markdown_cells.append(cell.source)

        markdown_content = "\n\n".join(markdown_cells)

        return markdown_content

    def _get_content_from_markdown_file(self) -> str:
        """Returns the content of a markdown file."""
        with open(self._content_file_path) as file:
            markdown_content = file.read()
        return markdown_content

    def _get_content(self) -> str:
        content = ""
        match self._content_file_type:
            case ContentType.MARKDOWN:
                content = self._get_content_from_markdown_file()
            case ContentType.NOTEBOOK:
                content = self._get_content_from_notebook_file()
        return content

    def _request_data(self) -> dict[str, str]:
        """Prepares the request data to be sent to the CMS endpoint."""
        return {
            "_id": self._get_content_id(),
            "displayName": self.display_name,
            "markdown": self._get_content(),
        }

    def sync_content(self) -> Response:
        """Sends the request to the CMS endpoint and returns the content id from the response."""
        response = self._request_handler.send_request(
            url=self._post_url, data=self._request_data()
        )

        if response.status_code != 200:
            raise ValueError(
                f"Request to the CMS failed with status code {response.status_code}."
            )
        if response.body is None:
            raise ValueError("Response body from CMS could not be parsed.")
        if self.ID_KEY not in response.body:
            raise ValueError(
                f"Response from the CMS does not contain the expected key '{self.ID_KEY}'."
            )
        result = response.body[self.ID_KEY]
        if not isinstance(result, str):
            raise ValueError(
                f"Response from the CMS does not contain a valid content id: {result}"
            )
        content_id: str = result
        self._save_content_id(content_id)

        return response



from dataclasses import dataclass
from typing import Any
from typing import Protocol

import requests


@dataclass
class Response:
    """The response object used in the package."""

    status_code: int
    body: dict[str, Any]


class RequestHandler(Protocol):
    """Interface for handling how a request are sent.

    Implementing classes may handle authentication, sessions, etc.
    """

    def send_request(
        self,
        url: str,
        headers: dict[str, str] | None = None,
        data: dict[str, str] | None = None,
    ) -> Response:
        """Sends the request to the specified url, optionally with headers and data, and returns the response."""
        ...


class BasicRequestHandler:
    """Basic, unauthenticated request handler."""

    def send_request(
        self,
        url: str,
        headers: dict[str, str] | None = None,
        data: dict[str, str] | None = None,
    ) -> Response:
        """Sends the request to the specified url without any headers."""
        response = requests.post(
            url,
            data=data,
        )

        try:
            body = response.json()
            body = dict(body)
        except Exception:
            body = {}

        return Response(
            status_code=response.status_code,
            body=body,
        )



from enum import Enum

from platformdirs import user_cache_path
from platformdirs import user_config_path
from platformdirs import user_data_path

APP_NAME = "pubmd"

METADATA_FILE = user_data_path(APP_NAME, ensure_exists=True) / "metadata.json"
CACHE_FILE = user_cache_path(APP_NAME, ensure_exists=True) / "cache.json"
CONFIG_FILE = user_config_path(APP_NAME, ensure_exists=True) / "config.json"

CACHE_FILE.touch()
CONFIG_FILE.touch()
METADATA_FILE.touch()


class ContentType(Enum):
    """Allowed content types."""

    MARKDOWN = ".md"
    NOTEBOOK = ".ipynb"



import json
from dataclasses import dataclass
from datetime import datetime

import jwt
import requests
from google.cloud import secretmanager

from ssb_pubmd.request_handler import Response

TYPE = "JWT"
ALGORITHM = "RS256"


@dataclass
class SecretData:
    """Data class to hold private key and connected data."""

    private_key: str
    kid: str
    principal_key: str


class JWTRequestHandler:
    """This class is used to send requests with a JSON Web Token (JWT) in the header."""

    def __init__(self, gc_secret_resource_name: str) -> None:
        """Initializes a JWT request handler object."""
        self._gc_secret_resource_name: str = gc_secret_resource_name

    def _private_key_from_secret_manager(self) -> SecretData:
        """Fetches the private key from Google Cloud Secret Manager."""
        client = secretmanager.SecretManagerServiceClient()
        print(f"Fetching secret from {self._gc_secret_resource_name}")
        response = client.access_secret_version(name=self._gc_secret_resource_name)
        raw_data = response.payload.data.decode("UTF-8")
        data = json.loads(raw_data)
        try:
            secret_data = SecretData(
                private_key=data["privateKey"],
                kid=data["kid"],
                principal_key=data["principalKey"],
            )
        except KeyError as e:
            raise ValueError(
                "The secret must be a JSON object with keys 'privateKey', 'kid' and 'principalKey'."
            ) from e
        return secret_data

    def _generate_token(self) -> str:
        secret_data = self._private_key_from_secret_manager()

        header = {
            "kid": secret_data.kid,
            "typ": TYPE,
            "alg": ALGORITHM,
        }

        iat = int(datetime.now().timestamp())
        exp = iat + 30
        payload = {
            "sub": secret_data.principal_key,
            "iat": iat,
            "exp": exp,
        }

        token = jwt.encode(
            payload, secret_data.private_key, algorithm=ALGORITHM, headers=header
        )
        return token

    def send_request(
        self,
        url: str,
        headers: dict[str, str] | None = None,
        data: dict[str, str] | None = None,
    ) -> Response:
        """Sends the request to the specified url with bearer token in header."""
        token = self._generate_token()
        headers = {
            "Authorization": f"Bearer {token}",
            "Content-Type": "application/json",
        }
        response = requests.post(
            url,
            headers=headers,
            json=data,
        )

        try:
            body = response.json()
            body = dict(body)
        except Exception:
            body = {}

        return Response(
            status_code=response.status_code,
            body=body,
        )



"""Command-line interface."""

import json
import os
from enum import Enum
from pathlib import Path
from typing import cast
from urllib.parse import urlparse

import click

from ssb_pubmd.browser_request_handler import BrowserRequestHandler
from ssb_pubmd.browser_request_handler import CreateContextMethod
from ssb_pubmd.constants import APP_NAME
from ssb_pubmd.constants import CACHE_FILE
from ssb_pubmd.constants import CONFIG_FILE
from ssb_pubmd.jwt_request_handler import JWTRequestHandler
from ssb_pubmd.markdown_syncer import MarkdownSyncer


class ConfigKey(Enum):
    """Configuration keys for the application."""

    BASE_URL = "base_url"
    LOGIN_URL = "login_url"
    POST_URL = "post_url"
    AUTH_METHOD = "auth_method"
    GC_SECRET_RESOURCE_NAME = "gc_secret_resource_name"


def get_config_value(config_key: ConfigKey) -> str:
    """Load a configuration value, with precedence environment variable > config file."""
    key = config_key.value

    def get_env_value() -> str:
        """Get value from environment variable, by uppercasing the key and adding prefix."""
        prefix = f"{APP_NAME.upper()}_"
        value = os.getenv(f"{prefix}{key.upper()}")

        return cast(str, value)

    def get_config_file_value() -> str:
        """Get value from the config file."""
        with open(CONFIG_FILE) as f:
            data = json.load(f)

        value = data.get(key)

        return cast(str, value)

    return get_env_value() or get_config_file_value()


def set_config_value(config_key: ConfigKey, value: str) -> None:
    """Set a configuration value in the config file."""
    key = config_key.value

    with open(CONFIG_FILE) as f:
        try:
            data = json.load(f)
        except json.JSONDecodeError:
            data = {}

    data[key] = value

    with open(CONFIG_FILE, "w") as f:
        json.dump(data, f, indent=4)


@click.group()
def cli() -> None:
    """Pubmd - a tool to sync markdown and notebook files to a CMS."""
    pass


@cli.command()
def settings() -> None:
    """Set the login and post URL for the CMS."""
    login_url = click.prompt("Enter the login URL", type=str)
    set_config_value(ConfigKey.LOGIN_URL, login_url)

    post_url = click.prompt("Enter the post URL", type=str)
    set_config_value(ConfigKey.POST_URL, post_url)

    click.echo(f"\nSettings stored in:\n{click.format_filename(CONFIG_FILE)}")


@cli.command()
def login() -> None:
    """Log in to the CMS application."""
    login_url = get_config_value(ConfigKey.LOGIN_URL)
    request_handler = BrowserRequestHandler(CACHE_FILE, login_url)

    method = CreateContextMethod.FROM_LOGIN
    with request_handler.new_context(method=method):
        click.echo("Logging in...")

    click.echo(f"\nBrowser context stored in:\n{CACHE_FILE}")


def sync_with_browser(content_file_path: str) -> None:
    """Sync a markdown or notebook file to the CMS."""
    login_url = get_config_value(ConfigKey.LOGIN_URL)
    request_handler = BrowserRequestHandler(CACHE_FILE, login_url)

    with request_handler.new_context() as context:
        post_url = get_config_value(ConfigKey.POST_URL)
        syncer = MarkdownSyncer(post_url, request_handler)

        syncer.content_file_path = Path(content_file_path)
        response = syncer.sync_content()

        click.echo("Content synced successfully.")

        path = response.body.get("previewPath", "")
        preview = urlparse(login_url)._replace(path=path).geturl()
        if preview:
            page = context.new_page()
            page.goto(preview)
            click.echo(f"Preview opened in new browser: {preview}")
            click.echo("Close the browser tab to finish.")
            page.wait_for_event("close", timeout=0)
        else:
            click.echo("No preview url found in the response.")


def sync_with_jwt(content_file_path: str) -> None:
    """Sync a markdown or notebook file to the CMS."""
    gc_secret_resource_name = get_config_value(ConfigKey.GC_SECRET_RESOURCE_NAME)
    request_handler = JWTRequestHandler(gc_secret_resource_name)

    post_url = get_config_value(ConfigKey.POST_URL)
    syncer = MarkdownSyncer(post_url, request_handler)

    syncer.content_file_path = Path(content_file_path)
    response = syncer.sync_content()

    click.echo("Content synced successfully.")

    preview_path = response.body.get("previewPath", "")
    if preview_path:
        base_url = get_config_value(ConfigKey.BASE_URL)
        preview = urlparse(base_url)._replace(path=preview_path).geturl()
        click.echo(f"Preview url found in the response: {preview}")
    else:
        click.echo("No preview url found in the response.")


@cli.command()
@click.argument("content_file_path", type=click.Path())
def sync(content_file_path: str) -> None:
    """Sync a markdown or notebook file to the CMS."""
    auth_method = get_config_value(ConfigKey.AUTH_METHOD)
    if auth_method == "browser":
        sync_with_browser(content_file_path)
    else:
        sync_with_jwt(content_file_path)


if __name__ == "__main__":
    cli()
